# -*- coding: utf-8 -*-
"""
@author: ftong
"""
import sys
import scipy.io
from obspy import UTCDateTime
from obspy.clients.fdsn import Client
from obspy import read_events
import numpy as np

try:    #make pandas optional as it is only used for reading csv files
    import pandas as pd
except:
    pd = lambda x: x

def read_mat(matlab_file, mag_label, **kwargs):
    
    mag = []
    time = []
    
    mat = scipy.io.loadmat(matlab_file)

    for i in range(0,len(mat['Catalog'][0])):
        
        # Extract magnitudes
        if mat['Catalog'][0][i][0]==mag_label:
            mag_frame = mat['Catalog'][0][i]
            mag = mag_frame[2].flatten()
            continue
    
        # Extract times
        if mat['Catalog'][0][i][0]=='Time':
            time_frame = mat['Catalog'][0][i]
            t1 = time_frame[2].flatten()
            datenums = np.array(t1)
            t2= pd.to_datetime(datenums-719529, unit='D') #convert from Matlab's datenum format to human readable format
            time = [UTCDateTime(str(t)) for t in t2] # convert to list of obspy UTCDateTime objects
            
    if len(mag)==0:
        print("Magnitude column not found in .mat file")
        sys.exit()

    return time, mag

def read_csv(csv_file,  datenum= True, **kwargs):
    
    # df = pd.read_csv(csv_file, engine="pyarrow", **kwargs) #use pyarrow for faster input
    df = pd.read_csv(csv_file, **kwargs)
    
    df.columns= df.columns.str.lower() # convert header to all lowercase letters
    header = df.columns
    
    # replace with common column names if present
    if 'time' in header:
        t_label = 'time'
    elif 't' in header:
        t_label = 't'

    if 'lon' in header:
        x_label = 'lon'
    elif 'long' in header:
        x_label = 'long'
    elif 'longitude' in header:
        x_label = 'longitude'
    elif 'x' in header:
        x_label = 'x'
        
    if 'lat' in header:
        y_label = 'lat'
    elif 'latitude' in header:
        y_label = 'latitude'
    elif 'y' in header:
        y_label = 'y'

    if 'depth' in header:
        z_label = 'depth'
    elif 'z' in header:
        z_label = 'z'

    if 'mag' in header:
        mag_label = 'mag'    
    elif 'ml' in header:
        mag_label = 'ml'
    elif 'magnitude' in header:
        mag_label = 'magnitude'

    t_raw = df[t_label].to_numpy()
    lat = df[y_label].to_numpy()
    lon = df[x_label].to_numpy()
    depth = df[z_label].to_numpy()
    mag = df[mag_label].to_numpy()
    
    if datenum:    
        timestamps = pd.to_datetime(t_raw-719529, unit='D') #convert from Matlab's datenum format to human readable format
        
    time = timestamps.to_numpy() # get time as numpy datetime64 objects
    time = [UTCDateTime(str(t)) for t in time] # convert to list of obspy UTCDateTime objects
    
    return time, lat, lon, depth, mag


def read_fdsn(service, save_to_file=None, file_format="QUAKEML", **kwargs):
       
    client = Client(service)    
    cat = client.get_events(**kwargs)
    
    if save_to_file != None:
        cat.write(filename=save_to_file, format=file_format, **kwargs)
    
    return extract_from_obspy_cat(cat)


def read_obspy(file, **kwargs):
    
    cat = read_events(file)
    
    return extract_from_obspy_cat(cat)


def extract_from_obspy_cat(cat):
    
    mag = np.array([event.preferred_magnitude().mag for event in cat])
    time = [event.preferred_origin().time for event in cat]
    lat = [event.preferred_origin().latitude for event in cat]
    lon = [event.preferred_origin().longitude for event in cat]
    depth = [event.preferred_origin().depth for event in cat]
    
    return time, lat, lon, depth, mag


# -*- coding: utf-8 -*-
"""
@author: ftong
"""
import numpy as np
import bisect #used for bisecting list of events by start time
from tdasha.window import Window
import sys

try:    #make dask optional
    from dask import delayed
    import dask
    from dask.diagnostics import ProgressBar # use Dask progress bar
except:
    dask = lambda x: x
    
try:    #make tqdm optional
    from tqdm import tqdm
except:
    tqdm = lambda x: x
    
def time_window_length(t_input, time_unit='day'):
        
    t = assure_order(t_input, chrono='forward')
    
    n_events = len(t) # number of events in the catalog
    t_start = t[0] # time of first event, assumes event list sorted by time
    t_end = t[-1] # time of last event
    time_span= t_end-t_start #length of data set between first and last event
    
    if time_unit == 'day':
        time_span = time_span / 86400 #convert time span from seconds to days
    elif time_unit == 'hour':
        time_span = time_span / 3600 #convert time span from seconds to hours
    elif time_unit == 'minute':
        time_span = time_span / 60 #convert time span from seconds to minutes

    activity_rate = n_events / time_span
    
    window_length = np.ceil(80 / activity_rate) # use window length such that there are 80 events per window
    
    return window_length,activity_rate

def assure_order(t, chrono, x=[None]):
    
    # t is a list of time values, x must be a numpy array
    
    # check ordering of time based on first two events and reverses it if necessary
    if chrono=="forward" and t[1]<t[0] or chrono=="backward" and t[1]>t[0]:
        t_new = list(reversed(t))
        
        if len(x)>1: # x provided, so reverse also x then return both t and x
            x_new = np.flip(x)
            return t_new, x_new
        
        else: # x is empty, return only t
            return t_new

    
    else: # ordering is already as desired based on test
    
        if len(x)>1: # x provided, so return both t and x
            return t, x
        
        else: # x is empty, return only t
            return t
        
def parse_time(t, t_unit, t_step_raw, win_size_raw=False):

    t_start = t[0]
    t_end = t[-1]

    # convert step and window size to seconds
    if t_unit == 'day':
        t_step = t_step_raw*86400
        win_size = win_size_raw*86400 
    elif t_unit == 'hour':
        t_step = t_step_raw*3600
        win_size = win_size_raw*3600 
    elif t_unit == 'minute':
        t_step = t_step_raw*60
        win_size = win_size_raw*60
    
    if not win_size_raw: 
        return t_start, t_end, t_step #CNE windows
    else:
        return t_start, t_end, t_step, win_size #CTL windows

# def comp_exc_prob_CNE(M_input, Mc, rum, t_input, t_unit, t_step_raw, t_period, win_size, M_max=None, verbose=False, ci = False, parallel = True, **kwargs):

#     t, M = assure_order(t_input, chrono='backward', x=M_input)
    
#     t_start, t_end, t_step = parse_time(t, t_unit, t_step_raw)
    
#     windows = np.arange(t_start,t_end,-t_step)
#     exc_pr = [] # exceedance probability for each window
#     exc_pr_ci = [] # confidence interval for exc_pr
#     tc = [] # timestamp of end of each window

#     # Iterate through all the windows, select desired number of events and compute exceedance probability
#     nwin = len(windows)

#     if parallel:
#         print("Processing windows in parallel using Dask")
#         pbar = ProgressBar()
#         pbar.register()
#         iter = windows
#     else:
#         print("Processing windows")
#         iter = tqdm(windows) # use tqdm progress bar
        
#     for idx, tt in enumerate(iter): # tt is the end time of each window
#         j = len(t) - bisect.bisect_right(t[::-1], tt)

#         # fetch the next win_size number of events after event t[j], event times and their magnitudes
#         T_sel = t[j:j+win_size] # event times for window
#         M_sel = M[j:j+win_size] # event magnitudes for window
        
#         if len(T_sel) < win_size:
#             print("Not enough events left to fill window starting at", tt)
#             break
        
#         tc.append(tt)
        
#         if parallel: #run in parallel using dask
#             win = delayed(Window)(Mc, rum, M_sel, T_sel, t_unit, t_step, t_period, **kwargs)
#             # exc_pr.append(delayed(win.exc_prob)()) 
#             exc_pr.append(win.exc_prob())
            
#             if ci: # estimate confidence interval
#                 exc_pr_ci.append(delayed(win.exc_prob_ci)(**kwargs))

#         else: #run without parallel computing
#             win = Window(Mc, rum, M_sel, T_sel, t_unit, t_step, t_period, **kwargs)
#             exc_pr.append(win.exc_prob())
            
#             if ci: # estimate confidence interval
#                 exc_pr_ci.append(win.exc_prob_ci(**kwargs))
                
#             if verbose==True:
#                 print("\nWindow", idx+1, "of", nwin)
#                 print("\tbandwidth:", round(win.h,3))
#                 print("\texceedance probability:", round(100*exc_pr[idx],2), "%")

#                 # print(exc_pr[idx])
#                 if ci==True:
#                     print("\tconfidence interval:", round(100*exc_pr_ci[idx][0],2), "to", round(100*exc_pr_ci[idx][1],2), "%")

    
#     if parallel:
#         exc_pr = dask.compute(*exc_pr, scheduler='processes')
#         exc_pr_ci = dask.compute(*exc_pr_ci, scheduler='processes')
#         # exc_pr = dask.compute(*exc_pr)
#         exc_pr = exc_pr[::-1] #because reverse is not working when using tuples with dask    
#         exc_pr_ci = exc_pr_ci[::-1]    
#     else: #reverse operation is faster than slicing
#         exc_pr.reverse()
#         exc_pr_ci.reverse()

#     tc.reverse()
    
#     print("Exceedance Probability computed from", tc[0], "to", tc[-1])
    
#     return exc_pr, exc_pr_ci, tc

def comp_exc_prob(M_input, Mc, rum, t_input, t_unit, t_step_raw, t_period, win_size_raw, win_method, M_max=None, verbose = False, ci = False, parallel = True, **kwargs):
    t, M = assure_order(t_input, chrono='backward', x=M_input)
    
    if win_method == 'CNE':
        t_start, t_end, t_step = parse_time(t, t_unit, t_step_raw)
        win_size = win_size_raw
    elif win_method == 'CTL':
        t_start, t_end, t_step, win_size = parse_time(t, t_unit, t_step_raw, win_size_raw)
    else:
        print("Abort! Please select either CNE or CTL method for windowing...")
        sys.exit()

    windows = np.arange(t_start,t_end,-t_step)
    exc_pr = [] # exceedance probability for each window
    exc_pr_ci = [] # confidence interval for exc_pr
    tc = [] # timestamp of end of each window
    nwin = len(windows)
    
    if parallel:
        print("Processing windows in parallel using Dask")
        pbar = ProgressBar()
        pbar.register()
        iter = windows
    else:
        print("Processing windows")
        iter = tqdm(windows) # use tqdm progress bar

    # Iterate through all the windows, select desired number of events and compute exceedance probability
    for idx, tt in enumerate(iter): # tt is the end time of each window

        if win_method == 'CNE':
            j = len(t) - bisect.bisect_right(t[::-1], tt)
    
            # fetch the next win_size number of events after event t[j], event times and their magnitudes
            T_sel = t[j:j+win_size] # event times for window
            M_sel = M[j:j+win_size] # event magnitudes for window
            
            if len(T_sel) < win_size:
                print("Not enough events left to fill window starting at", tt)
                break

        elif win_method == 'CTL':
            j1 = len(t) - bisect.bisect_right(t[::-1], tt)
            j2 = len(t) - bisect.bisect_right(t[::-1], tt-win_size)
            
            # fetch all events within the time window
            T_sel = t[j1:j2] # event times
            M_sel = M[j1:j2] # event magnitudes
            
            num_events = len(T_sel)
    
            if num_events<50:
                # print("Insufficient events at time window", tt)
                exc_pr.append(float("NaN")) # insufficient events, set to null
                exc_pr_ci.append([float("NaN"),float("NaN")]) # insufficient events, set to null
                tc.append(tt)
                continue            # move on to next window

        tc.append(tt)

        if parallel: #run in parallel using dask
            win = delayed(Window)(Mc, rum, M_sel, T_sel, t_unit, t_step, t_period, **kwargs)
            # exc_pr.append(delayed(win.exc_prob)()) 
            exc_pr.append(win.exc_prob())
            
            if ci: # estimate confidence interval
                exc_pr_ci.append(delayed(win.exc_prob_ci)(**kwargs))

        else: #run without parallel computing
            win = Window(Mc, rum, M_sel, T_sel, t_unit, t_step, t_period, **kwargs)
            exc_pr.append(win.exc_prob())
            
            if ci: # estimate confidence interval
                exc_pr_ci.append(win.exc_prob_ci(**kwargs))
                
            if verbose==True:
                print("\nWindow", idx+1, "of", nwin)
                print("\tbandwidth:", round(win.h,3))
                print("\texceedance probability:", round(100*exc_pr[idx],2), "%")

                # print(exc_pr[idx])
                if ci==True:
                    print("\tconfidence interval:", round(100*exc_pr_ci[idx][0],2), "to", round(100*exc_pr_ci[idx][1],2), "%")


    if parallel:
        exc_pr = dask.compute(*exc_pr, scheduler='processes')
        exc_pr_ci = dask.compute(*exc_pr_ci, scheduler='processes')
        # exc_pr = dask.compute(*exc_pr)
        exc_pr = exc_pr[::-1] #because reverse is not working when using tuples with dask    
        exc_pr_ci = exc_pr_ci[::-1]    
    else: #reverse operation is faster than slicing
        exc_pr.reverse()
        exc_pr_ci.reverse()

    tc.reverse()
    
    print("Exceedance Probability computed from", tc[0], "to", tc[-1])
    
    return exc_pr, exc_pr_ci, tc



# def comp_exc_prob_CTL(M_input, Mc, rum, t_input, t_unit, t_step_raw, t_period, win_size_raw, M_max=None, verbose = False, ci = False, parallel = True, **kwargs):
        
#     t, M = assure_order(t_input, chrono='backward', x=M_input)
    
#     t_start, t_end, t_step, win_size = parse_time(t, t_unit, t_step_raw, win_size_raw)
    
#     windows = np.arange(t_start,t_end,-t_step)
#     exc_pr = [] # exceedance probability for each window
#     exc_pr_ci = [] # confidence interval for exc_pr
#     tc = [] # timestamp of end of each window
        
#     # Iterate through all the windows, select desired number of events and compute exceedance probability
    
#     nwin = len(windows)

#     if parallel:
#         print("Processing windows in parallel using Dask")
#         pbar = ProgressBar()
#         pbar.register()
#         iter = windows
#     else:
#         print("Processing windows")
#         iter = tqdm(windows) # use tqdm progress bar

#     for idx, tt in enumerate(iter): # tt is the end time of each window
#         j1 = len(t) - bisect.bisect_right(t[::-1], tt)
#         j2 = len(t) - bisect.bisect_right(t[::-1], tt-win_size)
        
#         # fetch all events within the time window
#         T_sel = t[j1:j2] # event times
#         M_sel = M[j1:j2] # event magnitudes
        
#         num_events = len(T_sel)

#         if num_events<50:
#             # print("Insufficient events at time window", tt)
#             exc_pr.append(float("NaN")) # insufficient events, set to null
#             exc_pr_ci.append([float("NaN"),float("NaN")]) # insufficient events, set to null
#             tc.append(tt)
#             continue            # move on to next window
        

#         if parallel: #run in parallel using dask
#             win = delayed(Window)(Mc, rum, M_sel, T_sel, t_unit, t_step, t_period, **kwargs)
#             # exc_pr.append(delayed(win.exc_prob)())
#             exc_pr.append(win.exc_prob())
            
#             if ci: # estimate confidence interval
#                 exc_pr_ci.append(delayed(win.exc_prob_ci)(**kwargs))
        
#         else: #run without parallel computing

#             win = Window(Mc, rum, M_sel, T_sel, t_unit, t_step, t_period, **kwargs)
#             exc_pr.append(win.exc_prob())
        
#             if ci: # estimate confidence interval
#                 exc_pr_ci.append(win.exc_prob_ci(**kwargs))
                
#             if verbose==True:
#                 print("\nWindow", idx+1, "of", nwin)
#                 print("\tbandwidth:", round(win.h,3))
#                 print("\texceedance probability:", round(100*exc_pr[idx],2), "%")

#                 # print(exc_pr[idx])
#                 if ci==True:
#                     print("\tconfidence interval:", round(100*exc_pr_ci[idx][0],2), "to", round(100*exc_pr_ci[idx][1],2), "%")
        
#         tc.append(tt)
        
        

#         # break
    
#     if parallel:
#         exc_pr = dask.compute(*exc_pr, scheduler='processes')
#         exc_pr_ci = dask.compute(*exc_pr_ci, scheduler='processes')
#         # exc_pr = dask.compute(*exc_pr)
#         exc_pr = exc_pr[::-1] #because reverse is not working when using tuples with dask    
#         exc_pr_ci = exc_pr_ci[::-1]    
#     else: #reverse operation is faster than slicing
#         exc_pr.reverse()
#         exc_pr_ci.reverse()
    
#     tc.reverse()
    
#     print("Exceedance Probability computed from", tc[0], "to", tc[-1])
    
#     return exc_pr, exc_pr_ci, tc








#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa

__all__ = [
    "io",
    "compute",
    "window",
]

# use date and time as the version number    
# import datetime
# __version__ = datetime.datetime.now().strftime('%Y.%m.%d.%H%M')

# use github release tag as the version number
import os
try:
    __version__ = os.environ["GITHUB_REF_NAME"]
except:
    __version__ = "0.0"


# -*- coding: utf-8 -*-
"""
Created on Wed Nov  2 14:59:48 2022

@author: ftong
"""
import numpy as np
from KDEpy import FFTKDE
from scipy.stats import norm
from inspect import signature
from resample.bootstrap import confidence_interval, resample
from scipy.optimize import root_scalar
from adaptivekde import sskernel, ssvkernel

class Window:
    
    
    def __init__(self, mc, rum, m, t, t_unit, t_step, t_period, **kwargs):
        
        self.mc = mc
        self.rum = rum
        self.m = m
        self.t = t
        self.t_unit = t_unit
        self.t_step = t_step
        self.t_period = t_period
        self.time_span = self.get_time_span(self.t, self.t_unit)
        self.act_rate = self.activity_rate()
        self.mm = self.doubling(self.m, self.mc)
        self.h = self.bandwidth_estimator(self.mm, **kwargs)
    
    def activity_rate(self):
        
        m = self.m
        time_span = self.time_span
        num_events = len(m)
        
        act_rate = num_events / time_span
        return act_rate
    
    def get_time_span(self, t, t_unit):
        
        time_span = (t[0]-t[-1])  # time span of window in seconds
        
        if t_unit == 'day':
            time_span = time_span/86400
        elif t_unit == 'hour':
            time_span = time_span/3600
        elif t_unit == 'minute':
            time_span = time_span/60
        else:
            print("time unit not supported")
            exit
            
        return time_span
            
    def doubling(self, m, mc):
                
        xx = np.array([- m + 2*mc, m]) #doubling
        xx = xx.flatten() # make into 1D array
        xx.sort() #sort smallest to largest
        
        return xx
    
    
    def bandwidth_estimator(self, x, bw='ISJ', **kwargs):
        
        # use scipy to estimate bandwidth
        # from scipy.stats import gaussian_kde
        # if bandwidth=="scott":            
        #     kde = gaussian_kde(xx)
        #     h = kde.scotts_factor()
        # elif bandwidth == "silverman":
        #     kde = gaussian_kde(xx)
        #     h = kde.silverman_factor()

        # use statsmodels to estimate bandwidth
        # import statsmodels.api as sm
        # if type(bandwidth)==str: #need to estimate bandwidth
        #     kde = sm.nonparametric.KDEUnivariate(xx)
        #     kde.fit(bw=bandwidth)
        #     h = kde.bw
        # else:
        #     h = bandwidth
                
        # use KDEpy to estimate bandwidth
        if bw == 'ls':
            args = self.filter_args(kwargs, self.h_plat) #filter kwargs
            h = self.h_plat(x, **args)
        elif bw == 'ss':
            args = self.filter_args(kwargs, sskernel) #filter kwargs
            kde = sskernel(x, **args)
            h = kde[2]
        # elif bw == 'ssv': #don't know how to use multiple bandwidth values in a window yet.
        
        elif type(bw)==str: 
            args = self.filter_args(kwargs, FFTKDE) #filter kwargs
            h = FFTKDE(bw=bw, **args).fit(x).bw            
        elif type(bw)==float:
            h = bw # use user-provided value for bandwidth
        

        
        return h
    
    def bootstrap_ci(self, size=2000, **kwargs):
        
        # use pingouin for bootstrap
        # from pingouin import compute_bootci
        # CDF_CI = compute_bootci(M_sel, func=cdf_nonpar_ci, n_boot=2000, paired=False, method='cper', confidence=0.98)
        
        # use scipy for bootstrap
        # from scipy.stats import bootstrap
        # M_sel = (M_sel,) #convert array to sequence
        # CDF_CI = bootstrap(M_sel, cdf_nonpar_ci, confidence_level = 0.98, n_resamples=2000, method = 'bca', random_state=1)
                
        # use arch for bootstrap
        # from arch.bootstrap import IIDBootstrap
        # args1 = self.filter_args(kwargs, IIDBootstrap) #filter args by what function actually accepts
        # bs = IIDBootstrap(self.m, **args1)
        # args2 = self.filter_args(kwargs, bs.conf_int)        
        # ci = bs.conf_int(self.cdf_nonpar_ci, reps = size, method = 'bca', size = 0.98, **args2)
        # CDF_CI = ci.flatten()
        
        # use resample for bootstrap
        args1 = self.filter_args(kwargs, confidence_interval) #filter args by what confidence_interval function actually accepts
        args2 = self.filter_args(kwargs, resample) #because confidence_interval calls resample function
        args = args1 | args2 #merge dictionaries of arguments
        a, b = confidence_interval(self.cdf_nonpar_ci, self.m, size=size, **args)
        CDF_CI = [a,b]
        
        return CDF_CI
    
    def scaling(self, x, h):
        # EVALUATES A VECTOR OF SCALING FACTORS FOR THE NONPARAMETRIC ADAPTATIVE
        # ESTIMATION
        
        # x - the n-element column vector of data values
        # h - the optimal smoothing factor
        # ambd - the resultant n-element row vector of local scaling factors
                
        n = len(x)
        c = np.sqrt(2 * np.pi)
        gau = []
        for i in range(0,n):
            gau.append(sum(np.exp(- 0.5 * ((x[i] - x) / h) ** 2)) / c / n / h)
        
        g = np.exp(np.mean(np.log(gau)))
        ambd = np.sqrt(g / gau)
        return ambd        
        
    
    def cdf_nonpar_ci(self, M, m_max=None):    
        rum = self.rum
        Mc =  self.mc
        h = self.h
        x = self.doubling(M, Mc)
        ambd = self.scaling(x,h)        
        
        Fgau_rum = norm.cdf((rum - x) / ambd / h).mean()
        Fgau_Mc = norm.cdf((Mc - x) / ambd / h).mean()
        CDF_RUM = 2 * (Fgau_rum-Fgau_Mc)
        
        # upper limit to magnitude distribution specified
        if m_max != None: 
            Fgau_Mmax = norm.cdf((m_max - x) / ambd / h).mean()    
            mian = 2 * (Fgau_Mmax-Fgau_Mc)
            CDF_RUM = CDF_RUM / mian     
              
        return CDF_RUM
    
    def exc_prob(self, m_max=None):
        
        t_period = self.t_period #time period
        CDF_RUM = self.cdf_nonpar_ci(self.m, m_max)
        act_rate = self.activity_rate()
        exc_pr = 1 - np.exp(np.multiply(- act_rate * (1 - CDF_RUM),t_period))
        return exc_pr
    
    def exc_prob_ci(self, **kwargs):
        t_period = self.t_period
        act_rate = self.act_rate
        cdf_ci = self.bootstrap_ci(**kwargs)
        ci_upper = 1 - np.exp(np.multiply(- act_rate * (1 - cdf_ci[0]),t_period))
        ci_lower = 1 - np.exp(np.multiply(- act_rate * (1 - cdf_ci[1]),t_period))
        return [ci_lower, ci_upper]
    
    def filter_args(self, args, function): #filter arguments dictionary by list of arguments that the function actually accepts 
        key_list = signature(function).parameters
        filtered_args = {key: value for key, value in args.items() if key in key_list}      
        return filtered_args
    
    def funct(self, t, x):
        p2 = 1.41421356
        n = len(x)
        yy = np.zeros(n)
        for i in range(n):
            xij = (x - x[i])**2 / t**2
            y = np.exp(-xij / 4) * ((xij / 2 - 1) / p2) - 2 * np.exp(-xij / 2) * (xij - 1)
            yy[i] = np.sum(y)
        fct = np.sum(yy) - 2 * n
        return fct
    
    def h_plat(self, x, solver='brenth'):

        n = len(x)
        x = np.sort(x)
        sig = np.std(x, ddof=1)
        interval = np.array([0.1 * sig / n**0.2, 2 * sig / n**0.2])
        x1 = self.funct(interval[0], x)
        x2 = self.funct(interval[1], x)

        if x1 * x2 < 0:
            fun = lambda t: self.funct(t, x)
            sol = root_scalar(fun, method=solver, bracket=interval)
            h = sol.root
        else:
            y1 = 10 ** (np.floor(np.log10(interval[0]))) / 10
            x1 = self.funct(y1, x)
            x2 = self.funct(interval[0], x)
            if x1 * x2 < 0:
                interval[1] = interval[0]
                interval[0] = y1
                fun = lambda t: self.funct(t, x)
                sol = root_scalar(fun, method=solver, bracket=interval)
                h = sol.root
            else:
                y1 = 10 ** (np.floor(np.log10(interval[0]))) / 10
                dy = 0.05 * y1
                k = 0
                hh = []
                for i in np.arange(y1, interval[0], dy):
                    k += 1
                    hh.append(self.funct(i, x))
                k = np.argmin(np.abs(hh))
                h = y1 + (k - 1) * dy

        return h


